# -*- python -*-
# $RCSfile: layereditor.py,v $
# $Revision: 1.32 $
# $Author: langer $
# $Date: 2004/07/26 20:53:04 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import switchboard
from common import debug
from common import mainthread
from common.IO import display
from common.IO import gfxmanager
from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville

# LayerEditor menu items are defined *outside* the LayerEditor class,
# because some of the items (eg, New, Edit) need to exist even if
# there is no LayerEditor.

OOFMenuItem = oofmenu.OOFMenuItem

menu = mainmenu.OOF.addItem(OOFMenuItem('LayerEditor', secret=1))
layermenu = menu.addItem(OOFMenuItem('LayerSet', cli_only=1))

menu.addItem(OOFMenuItem('File'))
settingsmenu = menu.addItem(OOFMenuItem('Settings'))

# In these menu items, "window" is a string with the name
# of the window in it -- this name is exposed in the UI.
def newLayerSetCB(menuitem, window):
    openLayerEditor()
    layerEditor.newLayerSetCB(window)
    
layermenu.addItem(OOFMenuItem('New',
                             callback=newLayerSetCB,
                             params=[
            parameter.StringParameter('window')]))

def loadLayerCB(menuitem, window, layerNumber):
    openLayerEditor()
    layerEditor.loadLayerCB(window, layerNumber)
    
layermenu.addItem(OOFMenuItem('Edit',
                              callback=loadLayerCB,
                              params=[parameter.StringParameter('window'),
                                      parameter.IntParameter('layerNumber')]))


def displayedObjectCB(menuitem, category, object):
    openLayerEditor()
    layerEditor.displayedObjectCB(category, object)

layermenu.addItem(OOFMenuItem('DisplayedObject',
                              callback=displayedObjectCB,
                              params=[
    whoville.WhoClassParameter('category', condition=whoville.allClasses,
                               tip='What type of object is being displayed'),
    whoville.AnyWhoParameter('object',
                             tip='The name of the object being displayed')]))


def addMethodCB(menuitem, method):
    openLayerEditor()
    layerEditor.addMethodCB(method)
    
layermenu.addItem(OOFMenuItem('Add_Method',
                              callback=addMethodCB,
                              params=[display.DisplayMethodParameter('method')]
                              ))


def copyMethodCB(menuitem, layernumber):
    openLayerEditor()
    layerEditor.copyMethodCB(layernumber)

layermenu.addItem(OOFMenuItem('Copy_Method',
                              callback=copyMethodCB,
                              params=[parameter.IntParameter('layernumber')]))


def replaceMethodCB(menuitem, layernumber, method):
    openLayerEditor()
    layerEditor.replaceMethodCB(layernumber, method)

layermenu.addItem(OOFMenuItem('Replace_Method',
                              callback=replaceMethodCB,
                              params=[parameter.IntParameter('layernumber'),
                                      display.DisplayMethodParameter('method')]
                              ))


def deleteMethodCB(menuitem, layernumber):
    openLayerEditor()
    layerEditor.deleteMethodCB(layernumber)
        
layermenu.addItem(OOFMenuItem('Delete_Method',
                              callback=deleteMethodCB,
                              params=[parameter.IntParameter('layernumber')]))


def sendLayerCB(menuitem, window):
    openLayerEditor()
    layerEditor.sendLayerCB(window)

layermenu.addItem(OOFMenuItem('Send',
                              callback=sendLayerCB,
                              params=[parameter.StringParameter('window')]))

autoSend = 1
def setAutoSendCB(menuitem, value):
    global autoSend
    autoSend = value
settingsmenu.addItem(oofmenu.CheckOOFMenuItem(
    'AutoSend',
    value=autoSend,
    callback=setAutoSendCB,
    help="Send layers to graphics windows as soon as they're edited."))



class LayerEditor:
    def __init__(self):
        global menu
        self.menu = menu

        # The state of the editor is kept in a LayerSet, which is the
        # layer currently being edited.  Because it's possible to set
        # the WhoClass in the editor without specifying a Who, and
        # because a LayerSet only contains a Who, it's necessary to
        # keep a separate currentWhoClass variable.
        self.currentLayerSet = display.LayerSet(whoville.nobody)
        self.currentWhoClass = whoville.noclass

    def show(self): pass                # redefined in GUI mode

    def destroy(self):
        global layerEditor
        layerEditor = None

    def newLayerSetCB(self, window):
        switchboard.notify("deselect all gfxlayers")
        if window is not None:
            gfxwindow = gfxmanager.gfxManager.getWindow(window)
        else:
            gfxwindow = None
        self.initialize(gfxwindow)
        self.updateEditee()
        self.show()

    def loadLayerCB(self, window, layerNumber):
        gfxwindow = gfxmanager.gfxManager.getWindow(window)
        self.initialize(gfxwindow, layerNumber)

    def initialize(self, gfxwindow, layerNumber=None):
        self.currentLayerSet.destroy()
        # If no layer number is explicitly specified, use the selected
        # layer in the given graphics window, if any.
        if gfxwindow is None:
            selectedLayer = None
        else:
            if layerNumber is None:
                layerNumber = gfxwindow.selectedLayerNumber()
                if layerNumber is None:
                    selectedLayer = None
                else:
                    selectedLayer = gfxwindow.getLayer(layerNumber)
            else:
                selectedLayer = gfxwindow.getLayer(layerNumber)
        if selectedLayer:
            self.currentLayerSet = selectedLayer.layerset.clone()
        else:
            self.currentLayerSet = display.LayerSet() # new empty layer set
        self.currentWhoClass = self.currentLayerSet.who.getClass()
        self.updateEditee(selectedLayer)
        
    def addMethodCB(self, method):
        self.currentLayerSet.addMethod(method)
        self.updateMethodList(method)

    def copyMethodCB(self, layernumber):
        self.currentLayerSet.copyMethod(layernumber)
        self.updateMethodList()

    def replaceMethodCB(self, layernumber, method):
        self.currentLayerSet.replaceMethod(layernumber, method)
        self.updateMethodList()

    def deleteMethodCB(self, layernumber):
        self.currentLayerSet.doomMethod(layernumber)
        self.updateMethodList()

    def sendLayerCB(self, window):
        if self.ready():
            gfxwindow = gfxmanager.gfxManager.getWindow(window)
            gfxwindow.incorporateLayerSet(self.currentLayerSet)
#            switchboard.notify("redraw")
            gfxwindow.draw()

    def displayedObjectCB(self, category, object):
        self.changeDisplayedObject(category, object)

    # When called from a script in gui mode, this function changes the
    # state of the widget, which calls the script...  The widget's
    # callback is recursion proof, but this function would be called
    # twice without its own recursion suppression.
    recursionInhibitor = 0
    def changeDisplayedObject(self, category, object):
        if not LayerEditor.recursionInhibitor:
            LayerEditor.recursionInhibitor = 1
            self.currentWhoClass = whoville.getClass(category)
            if self.currentWhoClass is not None:
                try:
                    who = self.currentWhoClass[object]
                except KeyError:
                    who = None
            else:
                who = None
            self.currentLayerSet.changeWho(who)
            self.updateEditee()
            LayerEditor.recursionInhibitor = 0

    def updateEditee(self, selectedlayer=None):
        self.updateWho()
        self.updateMethodList(selectedlayer)
        
    def updateWho(self): pass           # redefined in LayerEditorGUI
    def updateMethodList(self, *args): pass # redefined in LayerEditorGUI

    def ready(self):
        return len(self.currentLayerSet) > 0


##########################
    
layerEditor = None                      # the one and only layer editor
        
def openLayerEditor(*args):
    global layerEditor
    if not layerEditor:
        layerEditor = mainthread.runBlock(newLayerEditor)
    layerEditor.show()

def newLayerEditor():
    # This function is redefined when layereditorGUI.py is loaded.
    return LayerEditor()
    
mainmenu.OOF.Windows.addItem(oofmenu.OOFMenuItem('Layer_Editor',
                                                 callback=openLayerEditor))

