# -*- python -*-
# $RCSfile: meshmenu.py,v $
# $Revision: 1.74 $
# $Author: reida $
# $Date: 2004/09/15 18:21:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import switchboard
from SWIG.engine import masterelement
from common import debug
from common import enum
from common import labeltree
from common import quit
from common import utils
from common import microstructure
from common.IO import automatic
from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville
from common.IO import datafile
from common.IO import microstructureIO
from engine import fieldinit
from engine import meshcrosssection
from engine import skeletoncontext
from engine import skeletonelement
from engine.IO import meshparameters
from engine.IO import skeletonIO, meshIO
from engine.IO import meshIO
import SWIG.engine.equation
import SWIG.engine.field
import SWIG.engine.solver
import engine.mesh
import types
import string

OOF = mainmenu.OOF
meshmenu = mainmenu.OOF.addItem(oofmenu.OOFMenuItem('Mesh', cli_only=1))

####################

# Menu arguments for the New Mesh command are automatically generated
# for each MasterElement geometry (triangular, quadrilateral, etc).
# These routines create the associated parameter names from the
# geometry, and conversely extract the geometry from the names.

_epdict = {}

def elementParamName(nsides):
    name = "element" + `nsides`
    _epdict[name] = nsides
    return name

def nsidesFromParamName(name):
    try:
        return _epdict[name]
    except KeyError:
        return None

# Look for an enclosing mesh parameter -- if not found, use the
# enclosing skeleton parameter.  Mesh copying needs the first case,
# new mesh construction needs the second.
def meshNameResolver(param, startname):
    if param.automatic():
        basename = 'mesh'
    else:
        basename = startname

    try:
        meshname = param.group['mesh'].value
    except IndexError:
        skelname = param.group['skeleton'].value
        skelpath = labeltree.makePath(skelname)
    else:
        skelpath = labeltree.makePath(meshname)[:-1]

    return engine.mesh.meshes.uniqueName(skelpath + [basename])

    
def newMesh(menuitem, name, skeleton, **kwargs):
    # Store the automatically generated masterelement type arguments
    # in a dictionary keyed by element geometry, in the form expected
    # by the realelement routines in the SkeletonElements.  The
    # arguments have names determined by elementParamName() (above)
    # and values that are instances of the element Enum classes
    # created by getMasterElementEnums() in masterelement_.py.  The
    # SkeletonElements, however, want a dictionary of actual
    # MasterElements, keyed by the number of sides.
    edict = {}                          # dict to pass to the Skeleton
    for ename, var in kwargs.items():
        n = nsidesFromParamName(ename)   # extract geometry from Parameter name
        if n is not None:
            edict[n] = masterelement.getMasterElementFromEnum(var)

    skelpath = labeltree.makePath(skeleton)
    skel = skeletoncontext.skeletonContexts[skelpath].getObject()
    # nm = utils.uniqueName(name, _meshnames)
    # nm = name # Should always be right -- clean up later.
    # _meshnames.append(nm)

    # nm = engine.mesh.meshes.uniqueName(skelpath + [nm])

    femesh = skel.femesh(edict, skeletonelement.SkeletonElement.realmaterial)
    mash = engine.mesh.meshes.add(skelpath+[name], femesh,
                                  skeleton=skel, elementdict=edict)

    # mash.add_to_skeleton(skel, edict)
    return mash

    # mush = engine.mesh.Mesh(name, skel, edict)
    # engine.mesh.meshes.add(skelpath + [name], mush)
    # return mush

newmeshcmd = meshmenu.addItem(
    oofmenu.OOFMenuItem('New', callback=newMesh,
                        threadable=oofmenu.THREADABLE,
                        params=
                        parameter.ParameterGroup(
    parameter.AutomaticNameParameter(
    'name', value=automatic.automatic,
    resolver=meshNameResolver,
    tip="Name of the new Mesh.  Use automatic selection, or type in a name"),
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts) ),
                        help="Create a new Mesh from a Skeleton."
    ))

def buildNewMeshCommand():
    # Add master element parameters to the newMesh command.  Make sure
    # to reuse the parameters that are already there, so that their
    # default values are retained.
    params = [newmeshcmd.get_arg('name'),
              newmeshcmd.get_arg('skeleton')]
    masterenumdict = masterelement.getMasterElementEnums()
    elist = masterenumdict.items()
    elist.sort()                        # in order of increasing nsides
    for nsides, enumclass in elist:
        pname = elementParamName(nsides)
        try:
            param = newmeshcmd.get_arg(pname)
        except TypeError:
            param = enum.EnumParameter(pname, enumclass,
                                       enumclass(enumclass.names[0]),
                                       tip="Element to replace %d-sided skeleton elements"%nsides)
        params.append(param)
    newmeshcmd.replace_args(params)

buildNewMeshCommand()
switchboard.requestCallback("new master element", buildNewMeshCommand)

#####################################

def renameMesh(menuitem, mesh, name):
    oldmeshpath = labeltree.makePath(mesh)
    themesh = engine.mesh.meshes[oldmeshpath]
    themesh.reserve()
    themesh.begin_writing()
    try:
        themesh.rename(name, exclude=oldmeshpath[-1])
    finally:
        themesh.end_writing()
        themesh.cancel_reservation()

meshmenu.addItem(oofmenu.OOFMenuItem('Rename',
                                     callback=renameMesh,
                                     threadable=oofmenu.THREADABLE,
                                     params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    parameter.StringParameter('name', '',
                              tip='New name for the mesh, in quotes')
    ],
                                     help="Rename a Mesh."))

#######################################

def deleteMesh(menuitem, mesh):
    meshobj = engine.mesh.meshes[mesh]
    ## Since the mesh is being destroyed, we have to acquire its lock,
    ## but we *don't* release the lock.  Nobody else should be doing
    ## anything to the mesh, ever. 
    meshobj.reserve()
    meshobj.begin_writing()
    meshobj.destroy()               # removes mesh from engine.mesh.meshes

meshmenu.addItem(oofmenu.OOFMenuItem('Delete',
                                     threadable=oofmenu.THREADABLE,
                                     callback=deleteMesh,
                                     params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes)
    ],
                                     help="Delete a Mesh."))

#######################################

def copyMesh(self, mesh, name,
             copyField, copyEquation, copyBndyCond):

    basemesh = engine.mesh.meshes[mesh]
    basemesh.begin_reading()
    try:
        edict = basemesh.elementdict
        copiedmeshname = name

        skel = basemesh.getSkeleton()
        skelpath = labeltree.makePath(basemesh.path())[:-1]

        copiedfemesh = skel.femesh(edict)
        newmesh = engine.mesh.meshes.add(skelpath+[copiedmeshname],
                                         copiedfemesh,
                                         skeleton=skel, elementdict=edict)
        newmesh.reserve()
        newmesh.begin_writing()
        try:
            copiedmesh = skelpath+[copiedmeshname]
            copiedmeshfullname = string.join(copiedmesh,":")
            basefemesh = basemesh.femesh()

            if copyField:
                for field in basemesh.all_fields():
                    copiedfemesh.define_field(field)
                    switchboard.notify("field defined",
                                       copiedmeshfullname, field.name(), 1)
                    copiedfemesh.acquire_field_data(field, basefemesh)
                    if basefemesh.is_active_field(field):
                        copiedfemesh.activate_field(field)
                        switchboard.notify("field activated",
                                           copiedmeshfullname, field.name(), 1)
                    if basefemesh.in_plane(field):
                        field.set_in_plane(copiedfemesh, 1)
                        switchboard.notify("field inplane",
                                           copiedmeshfullname, field.name(), 1)

            if copyEquation:
                for eqn in basemesh.all_equations():
                    copiedfemesh.activate_equation(eqn)
                    switchboard.notify('equation activated',
                                       copiedmeshfullname, eqn.name(), 1)

            if copyBndyCond:
                for (bcname, bc) in basemesh.allBoundaryConds():
                    copied = bc.copy(bc.boundary)
                    copied.add_to_mesh(bcname, copiedmesh)
        finally:
            newmesh.end_writing()
            newmesh.cancel_reservation()
    finally:
        basemesh.end_reading()


newmeshcmd = meshmenu.addItem(
    oofmenu.OOFMenuItem('Copy', callback=copyMesh,
                        threadable=oofmenu.THREADABLE,
                        params=
                        parameter.ParameterGroup(
    # whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts),
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    parameter.AutomaticNameParameter(
    'name', value=automatic.automatic, resolver=meshNameResolver,
    tip="Name of the copied Mesh. Use automatic selection, or type in a name"),
    parameter.BooleanParameter(
    'copyField', value=1, tip='copy field or not'),
    parameter.BooleanParameter(
    'copyEquation', value=1, tip='copy equation or not'),
    parameter.BooleanParameter(
    'copyBndyCond', value=1, tip='copy boundary conditions or not') ),
                        help="Copy a Mesh."
    ))

#######################################


# Copy the field state (definitions, active-ness, planarity) of one
# mesh into another.
def _copyFieldState(menuitem, source, target):
    source_mesh = engine.mesh.meshes[source]
    target_mesh = engine.mesh.meshes[target]
    source_mesh.begin_reading()
    target_mesh.reserve()
    target_mesh.begin_writing()
    try:
        source_obj = source_mesh.getObject()
        target_obj = target_mesh.getObject()
        source_fields = source_obj.all_fields()
        target_fields = target_obj.all_fields()

        if source_obj is target_obj:
            raise ooferror.ErrUserError('Source and target must differ!')

        # Undefine all the fields in the target that are not in the source.
        for f in target_fields:
            if not source_obj.is_defined_field(f):
                target_obj.undefine_field(f)
                switchboard.notify("field defined", target, f.name(), 0)

        for f in source_fields:
            # Definition.
            if not target_obj.is_defined_field(f):
                target_obj.define_field(f)
                switchboard.notify("field defined", target, f.name(), 1)

            # Activation.
            if source_obj.is_active_field(f):
                if not target_obj.is_active_field(f):
                    target_obj.activate_field(f)
                    switchboard.notify("field activated", target, f.name(), 1)
            else:
                if target_obj.is_active_field(f):
                    target_obj.deactivate_field(f)
                    switchboard.notify("field activated", target, f.name(), 0)

            # Planarity.
            if source_obj.in_plane(f):
                if not target_obj.in_plane(f):
                    f.set_in_plane(target_obj, 1)
                    switchboard.notify("field inplane", target, f.name(), 1)
            else:
                if target_obj.in_plane(f):
                    f.set_in_plane(target_obj, 0)
                    switchboard.notify("field inplane", target, f.name(), 0)

            try:
                initializer = source_mesh.initializers[f]
            except KeyError:
                pass
            else:
                target_mesh.initialize_field(f, initializer)
                switchboard.notify("field initialized")
    finally:
        source_mesh.end_reading()
        target_mesh.end_writing()
        target_mesh.cancel_reservation()
        
    engine.mesh.meshes[target].changed()
    

meshmenu.addItem(
    oofmenu.OOFMenuItem('Copy_Field_State',callback=_copyFieldState,
                        threadable=oofmenu.THREADABLE,
                        params=[
    whoville.WhoParameter('source',engine.mesh.meshes),
    whoville.WhoParameter('target',engine.mesh.meshes)],
                        help="Copy the Field state (defined, active, etc) from one Mesh to another."
                        ) )

    

# Likewise for equation state.
def _copyEquationState(menuitem, source, target):
    source_mesh = engine.mesh.meshes[source]
    target_mesh = engine.mesh.meshes[target]
    source_mesh.begin_reading()
    target_mesh.reserve()
    target_mesh.begin_writing()
    try:
        source_obj = source_mesh.getObject()
        target_obj = target_mesh.getObject()
        source_eqns = source_obj.all_equations()
        target_eqns = target_obj.all_equations()

        for e in target_eqns:
            if not source_obj.is_active_equation(e):
                target_obj.deactivate_equation(e)
                switchboard.notify("equation activated", target, e.name(), 0)

        for e in source_eqns:
            if not target_obj.is_active_equation(e):
                target_obj.activate_equation(e)
                switchboard.notify("equation activated", target, e.name(), 1)
    finally:
        source_mesh.end_reading()
        target_mesh.end_writing()
        target_mesh.cancel_reservation()
    engine.mesh.meshes[target].changed()
    

meshmenu.addItem(
    oofmenu.OOFMenuItem('Copy_Equation_State',callback=_copyEquationState,
                        threadable=oofmenu.THREADABLE,
                        params=[
    whoville.WhoParameter('source',engine.mesh.meshes),
    whoville.WhoParameter('target',engine.mesh.meshes)],
                        help="Copy the set of active Equations from one Mesh to another."
                        ) )


#######################################

fieldmenu = meshmenu.addItem(oofmenu.OOFMenuItem('Field'))


def _defineField(menuitem, mesh, field):
    # mesh and field are names, not objects
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        themesh = meshcontext.femesh()
        themesh.define_field(field)
        meshcontext.end_writing()
        switchboard.notify("field defined", mesh, field.name(), 1)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
def _undefineField(menuitem, mesh, field):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        meshcontext.femesh().undefine_field(field)
        switchboard.notify("field defined", mesh, field.name(), 0)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
    
fieldmenu.addItem(oofmenu.OOFMenuItem('Define',
                                      threadable=oofmenu.THREADABLE,
                                      callback=_defineField,
                                      params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.FieldParameter('field')
    ],
                                      help="Define a Field on a Mesh. Only defined Fields may be given values."))

fieldmenu.addItem(oofmenu.OOFMenuItem('Undefine',
                                      threadable=oofmenu.THREADABLE,
                                      callback=_undefineField,
                                      params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.FieldParameter('field')
    ],
                                      help="Undefine a Field on a Mesh.  Only defined Fields may be given values."))
                                        

def _activateField(menuitem, mesh, field):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        themesh = meshcontext.femesh()
        if themesh.is_defined_field(field):
            themesh.activate_field(field)
            switchboard.notify("field activated", mesh, field.name(), 1)
        else:
            reporter.report(
                "You must define a Field before you can activate it.")
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
    
def _deactivateField(menuitem, mesh, field):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        themesh = meshcontext.femesh()
        if themesh.is_active_field(field):
            themesh.deactivate_field(field)
            switchboard.notify("field activated", mesh, field.name(), 0)
        else:
            reporter.report(
                "You must define and activate a Field before you can deactivate it.")
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
    
fieldmenu.addItem(oofmenu.OOFMenuItem("Activate",
                                      threadable=oofmenu.THREADABLE,
                                      callback=_activateField,
                                      params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.FieldParameter('field')
    ],
                      help="Activate a Field on a Mesh.  The solver finds the values of active Fields."))

fieldmenu.addItem(oofmenu.OOFMenuItem('Deactivate',
                                      threadable=oofmenu.THREADABLE,
                                      callback=_deactivateField,
                                      params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.FieldParameter('field')
    ],
                      help="Deactivate a Field on a Mesh.  The solver finds the values of active Fields."))

def _inPlaneField(menuitem, mesh, field):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        themesh = meshcontext.femesh()
        field.set_in_plane(themesh, 1)
        switchboard.notify("field inplane", mesh, field.name(), 1)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
    
def _outOfPlaneField(menuitem, mesh, field):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        themesh = meshcontext.femesh()
        field.set_in_plane(themesh, 0)
        switchboard.notify("field inplane", mesh, field.name(), 0)    
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
    
fieldmenu.addItem(oofmenu.OOFMenuItem('In_Plane',
                                      threadable=oofmenu.THREADABLE,
                                      callback=_inPlaneField,
                                      params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.FieldParameter('field')
    ],
             help="In-plane Fields are constrained to have no z-components."))

fieldmenu.addItem(oofmenu.OOFMenuItem('Out_of_Plane',
                                      threadable=oofmenu.THREADABLE,
                                      callback=_outOfPlaneField,
                                      params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.FieldParameter('field')
    ],
                help="Out-of-plane Fields are allowed to have z-components."))

####################

eqnmenu = meshmenu.addItem(oofmenu.OOFMenuItem('Equation'))

def _activateEquation(menuitem, mesh, equation):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        themesh = meshcontext.femesh()
        themesh.activate_equation(equation)
        switchboard.notify('equation activated', mesh, equation.name(), 1)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
    
def _deactivateEquation(menuitem, mesh, equation):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        themesh = meshcontext.femesh()
        themesh.deactivate_equation(equation)
        switchboard.notify('equation activated', mesh, equation.name(), 0)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    engine.mesh.meshes[mesh].changed()
    
eqnmenu.addItem(oofmenu.OOFMenuItem('Activate',
                                    callback=_activateEquation,
                                    threadable=oofmenu.THREADABLE,
                                    params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.EquationParameter('equation')
    ],
             help="Activate an Equation on a mesh.  The Solver solves the active Equations."))

eqnmenu.addItem(oofmenu.OOFMenuItem('Deactivate',
                                    callback=_deactivateEquation,
                                    threadable=oofmenu.THREADABLE,
                                    params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.EquationParameter('equation')
    ],
             help="Activate an Equation on a mesh.  The Solver solves the active Equations."))

#######################################

csmenu = meshmenu.addItem(oofmenu.OOFMenuItem('Cross_Section'))

def csnameresolver(param, name):
    if param.automatic():
        basename = 'cs'
    else:
        basename = name
    meshname = param.group['mesh'].value
    if meshname is not None:
        meshpath = labeltree.makePath(meshname)
        meshobj = engine.mesh.meshes[meshpath]
        return meshobj.uniqueCSName(basename)

csnameparam = parameter.AutomaticNameParameter(
    'name', value=automatic.automatic, tip="Name this cross section.",
    resolver=csnameresolver)

csmeshparam = whoville.WhoParameter('mesh', engine.mesh.meshes,
                                    tip="Specify the mesh.")

csparamgroup = parameter.ParameterGroup(csnameparam, csmeshparam)

def _newCS(menuitem, mesh, name, cross_section):
    meshobj = engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.addCrossSection(name, cross_section)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()
    switchboard.notify("redraw")

csmenu.addItem(oofmenu.OOFMenuItem('New',
                                   threadable=oofmenu.THREADABLE,
                                   callback=_newCS,
                                   params=csparamgroup + [
    parameter.RegisteredParameter('cross_section',
                                  meshcrosssection.MeshCrossSection) ],
                                   help="Create a new cross section on a Mesh."))

def _delCS(menuitem, mesh, name):
    meshobj = engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.removeCrossSection(name)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()
    switchboard.notify("redraw")
        
csmenu.addItem(oofmenu.OOFMenuItem('Remove',
                                   threadable=oofmenu.THREADABLE,
                                   callback=_delCS,
                                   params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    parameter.StringParameter('name', tip='Cross section to remove')
    ],
                                   help='Delete a cross section from a mesh.'))

def _selectCS(menuitem, mesh, cross_section):
    meshobj = engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.selectCrossSection(cross_section)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()

    switchboard.notify("redraw")

csmenu.addItem(oofmenu.OOFMenuItem('Select',
                                   callback=_selectCS,
                                   params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    parameter.StringParameter('cross_section', tip='Cross section to select')
    ],
    help="Select a cross section on a mesh."))

def _renameCS(menuitem, mesh, cross_section, name):
    meshobj = engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.renameCrossSection(cross_section, name)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()

csmenu.addItem(oofmenu.OOFMenuItem('Rename',
                                   callback=_renameCS,
                                   params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    parameter.StringParameter('cross_section', tip='Cross section to rename'),
    parameter.StringParameter('name', tip='New name for the cross section')
    ],
    help="Rename a cross section on a mesh."))

def _editCS(menuitem, mesh, name, cross_section):
    meshobj = engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.replaceCrossSection(name, cross_section)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()
    switchboard.notify("redraw")

csmenu.addItem(oofmenu.OOFMenuItem('Edit',
                                   callback=_editCS,
                                   params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    parameter.StringParameter('name', tip='Cross section to edit'),
    parameter.RegisteredParameter('cross_section',
                                  meshcrosssection.MeshCrossSection,
                                  tip='New value for the cross section')
    ],
                                   help="Reparametrize a cross section on a mesh."))


#######################################

def initField(menuitem, mesh, field, initializer):
    # This routine is repeated verbatim in meshIO.py.
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        meshcontext.initialize_field(field, initializer)
        switchboard.notify("field initialized")
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    meshcontext.changed()
    
meshmenu.addItem(oofmenu.OOFMenuItem('Initialize',
                                     threadable=oofmenu.THREADABLE,
                                     callback = initField,
                                     params=[
    whoville.WhoParameter('mesh', engine.mesh.meshes),
    meshparameters.FieldParameter('field'),
    fieldinit.FieldInitParameter('initializer')
    ],
                                     help="Assign values to a Field on a Mesh.  The Field must be already defined."))


#####################

def saveMesh(menuitem, filename, mode, format, mesh):
    quit.disallow()
    try:
        meshcontext = engine.mesh.meshes[mesh]
        meshcontext.begin_reading()
        try:
            meshpath = labeltree.makePath(mesh)
            skelpath = meshpath[:2]
            skelcontext = skeletoncontext.skeletonContexts[skelpath]
            dfile = datafile.writeDataFile(filename, mode, format)
            microstructureIO.writeMicrostructure(
                dfile,
                microstructure.getMSContextFromMS(skelcontext.getObject().MS))
            skeletonIO.writeSkeleton(dfile, skelcontext)
            meshIO.writeMesh(dfile, meshcontext, skelcontext)
            dfile.close()
        finally:
            meshcontext.end_reading()
    finally:
        quit.allow()

OOF.File.Save.addItem(oofmenu.OOFMenuItem(
    'Mesh',
    callback = saveMesh,
    threadable=oofmenu.THREADABLE,
    params = [
    parameter.StringParameter('filename'),
    parameter.StringParameter('mode'),
    enum.EnumParameter('format', datafile.DataFileFormat, datafile.ASCII),
    whoville.WhoParameter('mesh', engine.mesh.meshes)],
    help="Save a Mesh to a file."
    ))

##########################
