# -*- python -*-
# $RCSfile: checkandmark.py,v $
# $Revision: 1.46 $
# $Author: haan $
# $Date: 2004/04/30 17:14:25 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from common import registeredclass
from common.IO import parameter
from common import enum
from common import utils
from SWIG.common.IO import progressbar

import math

class CheckAndMark(registeredclass.RegisteredClass):
    registry = []

    # Derived classes must override at least one of these functions:
    def preprocess(self, skeleton, divisions, markedEdges):
        pass

    def markSegment(self, segment, divisions, markedEdges):
        markedEdges.mark(segment.nodes()[0], segment.nodes()[1], divisions)

    def markElement(self, element, divisions, markedEdges):
        nnodes = element.nnodes()
        for i in range(nnodes):
            markedEdges.mark(element.nodes[i], element.nodes[(i+1)%nnodes],
                             divisions)
    
    def __call__(self, skeleton, context, element, divisions, markedEdges,
                 criterion):
        pass

#########################################

class CheckAllElements(CheckAndMark):     
    def __call__(self, skeleton, context, divisions, markedEdges, criterion):
        pBar = progressbar.getcProgressBar()
        pBar.set_message("marking all elements...                  ")
        elements = skeleton.activeElements()
        n = len(elements)
        for i in range(n):
            element = elements[i]
            if criterion(skeleton, element):
                self.markElement(element, divisions, markedEdges)
            if pBar.query_stop() :
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                pBar.set_progress(1.0*(i+1)/n)
                pBar.set_message("marking all elements: %d/%d" % (i+1, n))

registeredclass.Registration('All Elements', CheckAndMark,
                             CheckAllElements, 5.,
                             tip="Refine all elements.")

###########################

class CheckSelectedElements(CheckAndMark):
    def __call__(self, skeleton, context, divisions, markedEdges, criterion):
        pBar = progressbar.getcProgressBar()
        elements = context.elementselection.retrieve()
        n = len(elements)
        for i in range(n):
            element = elements[i]
            if element.active(skeleton) and \
               criterion(skeleton, element):
                self.markElement(element, divisions, markedEdges)
            if pBar.query_stop() :
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                pBar.set_progress(1.0*(i+1)/n)
                pBar.set_message("marking selected elements: %d/%d" % (i+1, n))

registeredclass.Registration('Selected Elements', CheckAndMark,
                             CheckSelectedElements, 0.2,
                             tip="Refine selected elements.")

######################

class CheckHomogeneity(CheckAndMark):
    def __init__(self, threshold):
        self.threshold = threshold
        
    def __call__(self, skeleton, context, divisions, markedEdges, criterion):
        pBar = progressbar.getcProgressBar()
        pBar.set_message("marking heterogeneous elements...                    ")
        elements = skeleton.activeElements()
        n = len(elements)
        for i in range(n):
            element = elements[i]
            if element.homogeneity(skeleton) < self.threshold and \
               criterion(skeleton, element):
                self.markElement(element, divisions, markedEdges)
            if pBar.query_stop() :
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                pBar.set_progress(1.0*(i+1)/n)
                pBar.set_message("marking heterogeneous elements: %d/%d" % (i+1, n))  
registeredclass.Registration('Heterogeneous Elements', CheckAndMark,
                             CheckHomogeneity, 0.1,
                             params=[
    parameter.FloatRangeParameter('threshold', (0.0, 1.0, 0.05), value=0.9,
        tip='refine elements whose homogeneity is less than this threshold')
                             ],
                             tip='Refine heterogeneous elements.')

#####################

class SegmentChooser(registeredclass.RegisteredClass):
    registry = []

class FromAllSegments(SegmentChooser):
    def getSegments(self, context):
        return context.getObject().activeSegments()

registeredclass.Registration('All Segments', SegmentChooser,
                             FromAllSegments, 0)

class FromSelectedSegments(SegmentChooser):
    def getSegments(self, context):
        segments = []
        for seg in context.segmentselection.retrieve():
            if seg.active(context.getObject()):
                segments.append(seg)
        return segments

registeredclass.Registration('Selected Segments', SegmentChooser,
                             FromSelectedSegments, 1)

class FromSelectedElements(SegmentChooser):
    def getSegments(self, context):
        segments = {}
        for elem in context.elementselection.retrieve():
            for i in range(elem.nnodes()):
                n0 = elem.nodes[i]
                n1 = elem.nodes[(i+1)%elem.nnodes()]
                seg = context.getObject().findSegment(n0, n1)
                if seg.active(context.getObject()):
                    segments[seg] = None
        return segments.keys()

registeredclass.Registration('Selected Elements', SegmentChooser,
                             FromSelectedElements, 2)    

class CheckHeterogeneousEdges(CheckAndMark):
    def __init__(self, threshold, choose_from):
        self.threshold = threshold
        self.choose_from = choose_from
        
    def __call__(self, skeleton, context, divisions, markedEdges, criterion):
        microstructure = skeleton.MS
        for segment in self.choose_from.getSegments(context):
            if segment.homogeneity(microstructure) < self.threshold:
                self.markSegment(segment, divisions, markedEdges)

registeredclass.Registration('Heterogeneous Segments', CheckAndMark,
                             CheckHeterogeneousEdges, 2,
                             params=[
    parameter.FloatRangeParameter('threshold', (0.0, 1.0, 0.05), value=0.9,
                                  tip="refine segments whose homogeneity is less than this threshold"),
    parameter.Parameter('choose_from', SegmentChooser)],
                             tip="Divide heterogeneous segements.")

class CheckSelectedEdges(CheckAndMark):
    def __call__(self, skeleton, context, divisions, markedEdges, criterion):
        for segment in context.segmentselection.retrieve():
            if segment.active(skeleton):
                self.markSegment(segment, divisions, markedEdges)

registeredclass.Registration('Selected Segments', CheckAndMark,
                             CheckSelectedEdges, 3,
                             tip="Divide selected segments.")

#####################

class CheckAspectRatio(CheckAndMark):
   def __init__(self, threshold):
       self.threshold = threshold
   def __call__(self, skeleton, context, divisions, markedEdges, criterion):
       for element in skeleton.activeElements():       
           if element.aspectRatio() > self.threshold:
               for segment in element.getLongSegments(skeleton):
                   if segment.active(skeleton):
                       self.markSegment(segment, divisions, markedEdges)

registeredclass.Registration('Aspect Ratio', CheckAndMark,
                            CheckAspectRatio, 4,
                            params=[
   parameter.FloatParameter('threshold', value=5.0,
                       tip="refine the long edges of elements whose aspect ratio is greater than this")],
                            tip="Divide elements with extreme aspect ratios.")
