# -*- python -*-
# $RCSfile: deputy.py,v $
# $Revision: 1.14 $
# $Author: langer $
# $Date: 2004/08/12 17:52:12 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# A DeputySkeleton is a lightweight way of representing a Skeleton
# which differs from a real Skeleton only in the positions of some of
# its nodes.  When the DeputySkeleton is being used as a Skeleton, it
# should be activated (with DeputySkeleton.activate()).  Only one
# DeputySkeleton of a given Skeleton can be active at one time.  When
# the DeputySkeleton is inactive, it's a mistake to move any of its
# nodes, and when it's active, it's a mistake to move any of the
# reference Skeleton's nodes.

# When a Deputy's Skeleton is about to be overwritten in the
# SkeletonContext's Ringbuffer, the Deputy must be promoted.  The
# oldest Deputy becomes the Sheriff, and all the other Deputies must
# become deputies of the new Sheriff.  The various DeputyTrackers
# associated with the promoted Deputy must also be promoted. 

from SWIG.common import timestamp
from common import debug
from engine import deputytracker
from engine import skeletonnode
import engine.skeleton

class DeputySkeleton(engine.skeleton.SkeletonBase):
    def __init__(self, skel, id = None):       # skel may be a deputy
        engine.skeleton.SkeletonBase.__init__(self)
        self.skeleton = skel.sheriffSkeleton()
        self.skeleton.addDeputy(self)
        self._illegal = skel._illegal
        # When the DeputySkeleton is active, the positions of its nodes
        # are stored in the actual nodes, and the positions of the
        # reference Skeleton's nodes are stored in the nodePositions
        # dictionary.  When it's inactive, it's the other way around.
        self.nodePositions = skel.getMovedNodes()
        self.active = 0
        self.timestamp = timestamp.TimeStamp()
        self.meshes = []
        self._destroyed = 0
        
        self.parallel_init(skel.name(), id)
        ## in parallel mode parallel_init(self) adds the object 
        ## to the ParallelObjectManager and sends the mirror
        ## deputy skeleton creation command and its ID to the
        ## back-end. The back-end creates a deputy and adds the ID
        ## to the back-end ParallelObjectManager

    def isDeputy(self):
        return 1

    def parallel_init(self, skel, _id):
        pass ## this is overwritten in parallel mode
    
    def deputyCopy(self):
        return DeputySkeleton(self)

    def getMovedNodes(self):
        nodedict = {}
        if self.active:                 # copy positions from Skeleton
            for node in self.nodePositions.keys():
                nodedict[node] = node.position()
        else:                           # copy positions from self
            nodedict.update(self.nodePositions)
        return nodedict

    def nodePosition(self, node):
        # Get the position of the node in this skeleton even if a
        # depty is active.
        if self.active:
            return node.position()
        try:
            return self.nodePositions[node]
        except KeyError:
            return self.skeleton.nodePosition(node)

    def originalPosition(self, node):
        try:
            return self.nodePositions[node]
        except KeyError:
            return node.position()

    def sheriffSkeleton(self):
        return self.skeleton

    def activate(self):
        if not self.active:
            self.active = 1
            self.skeleton.deputize(self)
            self.swapPositions()
    def deactivate(self):
        # called when the master Skeleton or another Deputy is activated
        if self.active:
            self.active = 0
            self.swapPositions()
    def swapPositions(self):
        for node, position in self.nodePositions.items():
            self.nodePositions[node] = node.position()
            node.unconstrainedMoveTo(position)
        self.skeleton.needsHash()
        
    # Any attributes not defined here are obtained from the underlying
    # skeleton.  This is *not* a derived class, though, because the
    # point is that objects of this class require very little
    # storage.
    def __getattr__(self, attr):
        return getattr(self.skeleton, attr)

    def disconnect(self):
        pass

    def moveNodeTo(self, node, position):
        if not self.nodePositions.has_key(node):
            # Storing old position
            self.nodePositions[node] = node.position()
        node.moveTo(position)

    def moveNodeBy(self, node, delta):
        if not self.nodePositions.has_key(node):
            self.nodePositions[node] = node.position()
        node.moveBy(delta)

    def moveNodeBack(self, node):
        node.moveBack()
        if node.position() == self.nodePositions[node]:
            del self.nodePositions[node]

    def destroy(self, context):
        self.skeleton.removeDeputy(self, context)
        del self.skeleton
        del self.nodePositions
        meshes = self.meshes[:]
        for mesh in meshes:
            mesh.destroy()
        self.meshes = []
        self._destroyed = 1

    def destroyed(self):
        return self._destroyed

    def getTimeStamp(self):
        return max(self.timestamp, self.skeleton.MS.getTimeStamp())

    def mapBoundary(self, bdy, skeleton, **kwargs):
        pass

    def newSelectionTracker(self, selectionset):
        return deputytracker.DeputySelectionTracker(
            selectionset.seniorTracker(self.skeleton))

    def newGroupTracker(self, groupset):
        return deputytracker.DeputyGroupTracker(
            groupset.getTracker(self.skeleton))

    def newPinnedNodeTracker(self):
        return DeputyPinnedNodeTracker(self)

    def promoteTrackers(self, context):
        context.promoteDeputyTrackers(self)

    def __repr__(self):
        return "DeputySkeleton(%d, %d)" % (id(self), id(self.skeleton))

# It will only take care of node move changes.
class DeputyProvisionalChanges:
    def __init__(self):
        self.movednodes = []  # list of MoveNode objects
        self.cachedDeltaE = None        # Energy difference
        self.cachedIllegal = None
        self.elements = None              # Elements involved
        
    class MoveNode:                     # nested class definition
        def __init__(self, node=None, position=None):
            self.node = node
            self.position = position

    def moveNode(self, node, position):
        self.movednodes.append(
            self.MoveNode(node=node, position=position))

    def elBefore(self):
        # elBefore and elAfter are the same for
        # DeputyProvisionalChanges because Deputies only move nodes,
        # they don't change element topology.
        return self.elAfter()

    def elAfter(self):
        if self.elements is None:
            eldict = {}
            for mvnode in self.movednodes:
                for nbr in mvnode.node.neighborElements():
                    eldict[nbr] = 1
            self.elements = eldict.keys()
        return self.elements

    def makeNodeMove(self, deputy):
        for mvnode in self.movednodes:
            deputy.moveNodeTo(mvnode.node, mvnode.position)

    def moveNodeBack(self, deputy):
        for mvnode in self.movednodes:
            deputy.moveNodeBack(mvnode.node)

    def illegal(self, deputy):
        # Will this change produce any illegal elements?
        if self.cachedIllegal is None:
            self.makeNodeMove(deputy)
            for element in self.elAfter():
                if element.illegal():
                    self.moveNodeBack(deputy)
                    self.cachedIllegal = 1
                    break
            else:
                self.moveNodeBack(deputy)
                self.cachedIllegal = 0
        return self.cachedIllegal

    def deltaE(self, deputy, alpha):
        # Return the change in energy per element if this move were to
        # be accepted.
        if self.cachedDeltaE is None:
            # Energy before the change
            oldE = 0.0
            for element in self.elBefore():
                oldE += element.energyTotal(deputy, alpha)
            oldE /= len(self.elBefore())
            # Move nodes accordingly to simulate the change
            self.makeNodeMove(deputy)
            # Energy after the change
            newE = 0.0
            for element in self.elAfter():
                newE += element.energyTotal(deputy, alpha)
            newE /= len(self.elAfter())
            # Move node back
            self.moveNodeBack(deputy)
            # Energy differnce due to the change
            self.cachedDeltaE = newE - oldE
        return self.cachedDeltaE

    def accept(self, deputy):
        for mvnode in self.movednodes:
            deputy.moveNodeTo(mvnode.node, mvnode.position)


class DeputyPinnedNodeTracker(skeletonnode.PinnedNodeTracker):
    # Unlike the other deputy trackers, the DeputyPinnedNodeTracker
    # has its own data and doesn't refer to another tracker.  That's
    # because the same node can be pinned in a skeleton and unpinned
    # in a deputy skeleton.  DeputyPinnedNodeTrackers and
    # PinnedNodeTrackers differ only in how the pinned state is
    # propagated from skeleton to skeleton.

    def pinDown(self, node, clist, where):
        node.pinDown(clist, where)
        
    def pinUp(self, node, plist, where):
        node.pinUp(plist, where)

    def unpinDown(self, node, clist, where):
        node.unpinDown(clist, where)

    def unpinUp(self, node, plist, where):
        node.unpinUp(plist, where)

    def implied_pin(self, oldtracker):
        for n in oldtracker.get():
            there = oldtracker.nodePosition(n)
            here = self.nodePosition(n)
            if here == there:
                self.add(n)

    def __repr__(self):
        return "DeputyPinnedNodeTracker"

from common import parallel_enable
if parallel_enable.query() == 1:
    from engine import deputyParallel
    DeputySkeleton.parallel_init = deputyParallel._parallel_init
